#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <time.h>

#define PI 3.1415926535897932384626433832795

#include "ptc/tinyptc.h"
#include <bass/bass.h>

#define WIDTH 640
#define HEIGHT 480
#define SIZE WIDTH*HEIGHT

#define rnd(n) (rand()%(n+1))

static int screen[SIZE];

int vmem[160*240];
int pal[16];
int map[256*256];
int sprites[8][64*128];

struct Column
{
	int height;
	float x, y;
	int color[8];
};
Column columns[8];

float dx[3];	
float dy[3];

HSAMPLE		music_sample;
HCHANNEL	music_channel;
HSAMPLE		sound_sample[6];
HCHANNEL	sound_channel[6];

void play_sound(int i, int vol)
{
	if(sound_channel[i]==NULL)
		sound_channel[i]=BASS_SampleGetChannel(sound_sample[i], false);
	BASS_ChannelSetAttributes(sound_channel[i], 44100, vol, 0);
	BASS_ChannelPlay(sound_channel[i], true);
}

void clear_vmem()
{
	for(int i=0;i<160*240;i++)
		vmem[i]=0;
}

void update_screen()
{
	int di=0, si=0;
	int row[640];
	for(int y=0;y<240;y++)
	{
		for(int x=0;x<160;x++)
		{
			int c=pal[vmem[si++]];
			int ri=x*4;
			row[ri++]=c;
			row[ri++]=c;
			row[ri++]=c;
			row[ri++]=c;
		}
		for(int x=0;x<640;x++)
			screen[di++]=row[x];
		for(int x=0;x<640;x++)
			screen[di++]=row[x];
	}
}

void load_raw(int *dest, char *filename, int w, int h)
{
	FILE *file=fopen(filename, "rb");
	int di=0;
	for(int y=0;y<h;y++)
		for(int x=0;x<w;x++)
		{
			int c=0x000000;
			fread(&c, 3, 1, file);
			c=(c&0x00FF00)|((c&0xFF0000)>>16)|((c&0x0000FF)<<16);
			for(int ci=0;ci<16;ci++)
				if(c==pal[ci])
					break;
			if(ci==16)
				dest[di++]=-1; // transparent
			else
				dest[di++]=ci;
		}
	fclose(file);
}

void draw_spherical(int *patch, int sx, int sy, int w, int h, int ox, int oy, float scale, float dx, float dy)
{
	float ixscale=2.0f/scale;
	float iyscale=1.0f/scale;
	for(int y=sy;y<sy+h;y++)
	{
		int di=y*160+sx;
		for(int x=sx;x<sx+w;x++)
		{
			float px=(float)(x-ox)*ixscale;
			float py=(float)(y-oy)*iyscale;
			float dist=(float)sqrt(px*px+py*py);
			if(dist>=20.0f)
			{
				di++;
				continue;
			}
			float stretch=1.0f;
			if(dist>0.0f)
				stretch=asin(dist/20)*20/dist;
			px*=stretch*2.0f;
			py*=stretch*2.0f;
			px+=dx;
			py+=dy;
			px+=32.0f;
			py+=32.0f;
			int ipx=(int)px;
			int ipy=(int)py;
			ipx&=255;
			ipy&=255;
/*			if(ipx<0 || ipx>=64 || ipy<0 || ipy>=64)
			{
//				vmem[di++]=2;
				di++;
				continue;
			}*/
			vmem[di++]=patch[ipy*256+ipx];
		}
	}
}

void draw_sprite(int *sprite, int sx, int sy, int w, int h)
{
	for(int y=0;y<h;y++)
	{
		int di=(y+sy)*160+sx;
		int si=y*w;
		for(int x=0;x<w;x++)
		{
			int c=sprite[si++];
			if(c>-1)
				vmem[di++]=c;
			else
				di++;
		}
	}
}

void draw_zoomed(int *sprite, int sx, int sy, int w, int h, float zoom)
{
	int sw=(int)((float)w*zoom);
	int sh=(int)((float)h*zoom);
	int ox=(int)(zoom/2);
	int oy=(int)(zoom/2);
	for(int y=0;y<sh;y++)
	{
		int di=(y+sy-sh/2+oy)*160+sx-sw/2+ox;
		for(int x=0;x<sw;x++)
		{
			float sx=(float)(x-(float)sw/2)/sw*w+w/2;
			float sy=(float)(y-(float)sh/2)/sh*h+h/2;
			int isx=(int)sx;
			int isy=(int)sy;
			if(isx<0 || isx>=w || isy<0 || isy>=h)
			{
				di++;
				continue;
			}
			int c=sprite[isy*w+isx];
			if(c>-1)
				vmem[di++]=c;
			else
				di++;
		}
	}
}

void draw_surface_sprite(int *sprite, float bzx, float bzy, float altitude)
{
	float vx=(bzx-dx[0]*2);
	if(vx<-256) vx+=512;
	if(vx>256) vx-=512;
	float vy=(bzy-dy[0]*2);
	if(vy<-256) vy+=512;
	if(vy>256) vy-=512;
	vx=vx/5.6f;
	vy=vy/5.6f;
	float dist=(float)sqrt(vx*vx+vy*vy);
	float scale=(float)sin(dist/20*1.57f)*20;
	if(dist<20)
	{
		float bx=vx/dist*scale*5.6f*(1.05f+altitude);
		float by=vy/dist*scale*5.6f*(1.05f+altitude);
		if(by<90 && by>-90)
			draw_zoomed(sprite, 70+(int)bx/2, 120+(int)by, 8, 8, 1.5f-scale/30);
	}
}

void reset_columns()
{
	for(int i=0;i<8;i++)
		columns[i].height=0;

	for(int i=0;i<3;i++)
	{
		columns[i].height=1;
		columns[i].x=(float)rnd(511)-255;
		columns[i].y=(float)rnd(511)-255;
		columns[i].color[0]=i;
	}
}

void main()
{
	srand(time(NULL));

/*	pal[1]=0xEE33DD;
	pal[2]=0x2244AA;
	pal[3]=0x55CC22;*/
//	pal[2]=0x2244AA;
//	pal[3]=(255<<16)|(224<<8)|(150);

	for(int i=0;i<16;i++)
		pal[i]=0x1000000;

	pal[0]=0x000000;
	pal[1]=(255<<16)|(72<<8)|(0);
	pal[2]=(86<<16)|(128<<8)|(137);
	pal[3]=(192<<16)|(189<<8)|(180);

	pal[4]=(40<<16)|(0<<8)|(0);
	pal[5]=(20<<16)|(10<<8)|(80);
	pal[6]=(40<<16)|(50<<8)|(80);
	pal[7]=(10<<16)|(20<<8)|(60);

	pal[8]=(100<<16)|(255<<8)|(40);
	pal[9]=(120<<16)|(190<<8)|(255);
	pal[10]=(255<<16)|(224<<8)|(150);
	pal[11]=(255<<16)|(129<<8)|(190);

	HWND hwnd=(HWND)0xA4A4A4A4;
	HINSTANCE hinst=(HINSTANCE)0xA4A4A4A4;

	if(!ptc_open("LD", WIDTH, HEIGHT, hwnd, hinst))
		return;

	load_raw(map, "data\\map2.raw", 256, 256);
	for(int y=0;y<256;y++)
		for(int x=128;x<256;x++)
			map[y*256+x]+=4;

	load_raw(sprites[0], "data\\blue.raw", 8, 8);
	load_raw(sprites[1], "data\\orange.raw", 8, 8);
	load_raw(sprites[2], "data\\pink.raw", 8, 8);
	load_raw(sprites[3], "data\\patchling.raw", 64, 128);
	load_raw(sprites[4], "data\\patchlings.raw", 64, 128);
	load_raw(sprites[5], "data\\aim.raw", 32, 32);

	BASS_Init(1, 44100, 0, hwnd, NULL);
	BASS_Start();
	music_sample=BASS_SampleLoad(FALSE, "data\\ldmoon.ogg", 0, 0, 1, BASS_SAMPLE_OVER_POS);
	music_channel=BASS_SampleGetChannel(music_sample, false);
	BASS_ChannelSetAttributes(music_channel, 44100, 30, 0);
	BASS_ChannelSetFlags(music_channel, BASS_SAMPLE_LOOP);
	BASS_ChannelPlay(music_channel, true);
	for(int i=0;i<6;i++)
		sound_channel[i]=NULL;
	sound_sample[0]=BASS_SampleLoad(FALSE, "data\\launch.ogg", 0, 0, 1, BASS_SAMPLE_OVER_POS);
	sound_sample[1]=BASS_SampleLoad(FALSE, "data\\hit.ogg", 0, 0, 1, BASS_SAMPLE_OVER_POS);
	sound_sample[2]=BASS_SampleLoad(FALSE, "data\\column.ogg", 0, 0, 1, BASS_SAMPLE_OVER_POS);
	sound_sample[3]=BASS_SampleLoad(FALSE, "data\\miss.ogg", 0, 0, 1, BASS_SAMPLE_OVER_POS);
	sound_sample[4]=BASS_SampleLoad(FALSE, "data\\fail.ogg", 0, 0, 1, BASS_SAMPLE_OVER_POS);
	sound_sample[5]=BASS_SampleLoad(FALSE, "data\\win.ogg", 0, 0, 1, BASS_SAMPLE_OVER_POS);

	reset_columns();
	
	float limit[3];
	limit[2]=0.5f;
	limit[1]=5.0f;

	for(int i=0;i<4;i++)
	{
		dx[i]=0.0f;
		dy[i]=0.0f;
	}
	dx[1]=0.25f;
	dx[0]=12.0f;

	float zoom=0.0f;
	float dzoom=0;
	
	int launch=0;
	int current=rnd(2);
	int smirk=0;
	bool failed=false;
	
	char cheat[10];
	char cheat1[10]="SHAME";
	char cheat2[10]="SHOWYOU";
	char hkey=0;
	bool wimpy=false;
	bool vis=false;

	for(;;)
	{
		// logic
		
		if(GetAsyncKeyState(VK_ESCAPE)&KF_UP)
			break;
			
		for(int i=0;i<2;i++)
		{
			dx[i]+=dx[i+1];
			if(i>0)
			{
				if(dx[i]>limit[i]) dx[i]=limit[i];
				if(dx[i]<-limit[i]) dx[i]=-limit[i];
			}
			dy[i]+=dy[i+1];
			if(i>0)
			{
				if(dy[i]>limit[i]) dy[i]=limit[i];
				if(dy[i]<-limit[i]) dy[i]=-limit[i];
			}
		}
		if(dx[0]<-256) dx[0]+=512;
		if(dx[0]>256) dx[0]-=512;
		if(dy[0]<-256) dy[0]+=512;
		if(dy[0]>256) dy[0]-=512;
		if(smirk<1000)
		{
			if(wimpy)
			{
				dx[2]=0.0f;
				dy[2]=0.0f;
				if(GetAsyncKeyState(VK_RIGHT)&KF_UP)
					dx[1]+=0.01f;
				if(GetAsyncKeyState(VK_LEFT)&KF_UP)
					dx[1]-=0.01f;
				if(GetAsyncKeyState(VK_DOWN)&KF_UP)
					dy[1]+=0.01f;
				if(GetAsyncKeyState(VK_UP)&KF_UP)
					dy[1]-=0.01f;
			}
			else
			{
				if(GetAsyncKeyState(VK_RIGHT)&KF_UP)
					dx[2]+=0.0001f;
				if(GetAsyncKeyState(VK_LEFT)&KF_UP)
					dx[2]-=0.0001f;
				if(GetAsyncKeyState(VK_DOWN)&KF_UP)
					dy[2]+=0.0001f;
				if(GetAsyncKeyState(VK_UP)&KF_UP)
					dy[2]-=0.0001f;
			}

			if((GetAsyncKeyState('Z')&KF_UP) && launch==0 && dzoom>=150)
			{
				launch=1;
				play_sound(0, 20);
			}
		}

		if(launch>0)
		{
			launch++;
			if(launch>70)
			{
				int i;
				for(i=0;i<8;i++)
				{
					float sx=columns[i].x-dx[0]*2;
					if(sx<-256) sx+=512;
					if(sx>256) sx-=512;
					float sy=columns[i].y-dy[0]*2;
					if(sy<-256) sy+=512;
					if(sy>256) sy-=512;
					if(columns[i].height>0 && fabs(sx)<6.0f && fabs(sy)<6.0f)
					{
						if(columns[i].color[columns[i].height-1]==current)
						{
							failed=true;
							dzoom=149;
							dx[2]=0.05f;
							pal[0]=0x400000;
							BASS_ChannelStop(music_channel);
							play_sound(4, 20);
							play_sound(1, 20);
							break;
						}
						columns[i].height++;
						if(columns[i].height==8)
						{
							smirk=80;
							columns[i].height=0;
							int g;
							for(g=0;g<8;g++)
								if(columns[g].height>0)
									break;
							if(g==8)
							{
								pal[0]=0x103070;
								smirk=1000;
								dx[2]=0.02f;
								dy[2]=0.005f;
								BASS_ChannelStop(music_channel);
								music_channel=BASS_SampleGetChannel(sound_sample[5], false);
								BASS_ChannelSetAttributes(music_channel, 44100, 10, 0);
								BASS_ChannelSetFlags(music_channel, BASS_SAMPLE_LOOP);
								BASS_ChannelPlay(music_channel, true);
							}
							play_sound(2, 20);
						}
						else
						{
							columns[i].color[columns[i].height-1]=current;
							play_sound(1, 20);
						}
						break;
					}
				}
				if(i==8)
				{
					int ci;
					for(ci=0;ci<8;ci++)
						if(columns[ci].height==0)
						{
							columns[ci].height=1;
							columns[ci].x=(float)rnd(511)-255;
							columns[ci].y=(float)rnd(511)-255;
							columns[ci].color[0]=current;
							break;
						}
					if(ci==8)
					{
						failed=true;
						dzoom=149;
						dx[2]=0.05f;
						pal[0]=0x400000;
						BASS_ChannelStop(music_channel);
						play_sound(4, 20);
					}
					play_sound(3, 20);
				}
				launch=0;
				int col=-1;
				for(int g=0;g<8;g++)
					if(columns[g].height>0)
					{
						if(col==-1)
							col=columns[g].color[columns[g].height-1];
						else if(col!=columns[g].color[columns[g].height-1])
						{
							col=-1;
							break;
						}
					}
				current=col;
				while(current==col)
					current=rnd(2);
			}
		}
		
		if(smirk>0 && smirk<1000)
			smirk--;

		bool keyhit=false;
		for(char ch='A';ch<='Z';ch++)
			if(GetAsyncKeyState(ch)&KF_UP)
			{
				if(ch!=hkey)
				{
					for(int i=0;i<9;i++)
						cheat[i]=cheat[i+1];
					cheat[9]=ch;
					for(int i=0;i<5;i++)
						if(cheat[9-4+i]!=cheat1[i])
							break;
					if(i==5)
						wimpy=true;
					for(int i=0;i<7;i++)
						if(cheat[9-6+i]!=cheat2[i])
							break;
					if(i==7)
						vis=true;
				}
				hkey=ch;
				keyhit=true;
				break;
			}
		if(!keyhit)
			hkey=0;

		// gfx

		clear_vmem();

		if(dzoom<150)
		{
			if(failed)
			{
				dzoom--;
				zoom-=0.3f/(1.0f+sqrt((float)dzoom));
				if(dzoom==0)
				{
					for(int i=0;i<4;i++)
					{
						dx[i]=0.0f;
						dy[i]=0.0f;
					}
					dx[1]=0.25f;
					dx[0]=12.0f;

					reset_columns();

					failed=false;
					pal[0]=0x000000;
					zoom=0.0f;
					zoom+=0.3f/(1.0f+sqrt((float)dzoom));
					dzoom=1;
					music_channel=BASS_SampleGetChannel(music_sample, false);
					BASS_ChannelSetAttributes(music_channel, 44100, 30, 0);
					BASS_ChannelSetFlags(music_channel, BASS_SAMPLE_LOOP);
					BASS_ChannelPlay(music_channel, true);
				}
			}
			else
			{
				dzoom++;
				zoom+=0.3f/(1.0f+sqrt((float)dzoom));
			}
		}
		draw_spherical(map, 0, 30, 160, 180, 70, 120, zoom, dx[0], dy[0]);

		if(dzoom>=150)
		{
			for(int i=0;i<8;i++)
				for(int c=0;c<columns[i].height;c++)
					draw_surface_sprite(sprites[columns[i].color[c]], columns[i].x, columns[i].y, (float)c*0.05f);
		}

		if(smirk<1000)
			draw_zoomed(sprites[current], 70, 180-(int)(sin((float)launch/50*1.57)*70), 8, 8, 3.0f-(float)launch/70*1.5f);
		draw_sprite(sprites[5], 70-8, 120-16, 32, 32);

		int di1=0;
		int di2=210*160;
		for(int y=0;y<30;y++)
			for(int x=0;x<160;x++)
			{
				vmem[di1++]=0;
				vmem[di2++]=0;
			}
		for(int y=0;y<240;y++)
		{
			di1=y*160+140;
			for(int x=140;x<160;x++)
				vmem[di1++]=0;
		}

		if(vis)
		{		
			vmem[50*160+130]=1;
			for(int i=-100;i<100;i++)
			{
				if(i<0 && (float)i>dx[2]*4000)
					vmem[50*160+130+i]=2;
				if(i>0 && (float)i<dx[2]*4000)
					vmem[50*160+130+i]=2;
				if(i<0 && (float)i>dy[2]*8000)
					vmem[(50+i)*160+130]=2;
				if(i>0 && (float)i<dy[2]*8000)
					vmem[(50+i)*160+130]=2;
			}
		}

		draw_sprite(sprites[3+(int)(smirk>0)], 160-65, 240-130, 64, 128);

		update_screen();
		Sleep(10);
		ptc_update(screen);
	}

	BASS_Free();
	ptc_close();
	ExitProcess(0);
}

